#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from django.urls import reverse
from openstack_dashboard import api

from blazar_dashboard import api as blazar_api
from blazar_dashboard.test import helpers as test

import logging
LOG = logging.getLogger(__name__)

INDEX_TEMPLATE = 'admin/hosts/index.html'
INDEX_URL = reverse('horizon:admin:hosts:index')
DETAIL_TEMPLATE = 'admin/hosts/detail.html'
DETAIL_URL_BASE = 'horizon:admin:hosts:detail'
CREATE_URL = reverse('horizon:admin:hosts:create')
CREATE_TEMPLATE = 'admin/hosts/create.html'
UPDATE_URL_BASE = 'horizon:admin:hosts:update'
UPDATE_TEMPLATE = 'admin/hosts/update.html'


class HostsTests(test.BaseAdminViewTests):
    @mock.patch.object(blazar_api.client, 'host_list')
    def test_index(self, host_list):
        hosts = self.hosts.list()
        host_list.return_value = hosts

        res = self.client.get(INDEX_URL)

        host_list.assert_called_once_with(test.IsHttpRequest())
        self.assertTemplateUsed(res, INDEX_TEMPLATE)
        self.assertNoMessages(res)
        self.assertContains(res, 'compute-1')
        self.assertContains(res, 'compute-2')

    @mock.patch.object(blazar_api.client, 'host_list')
    def test_index_no_hosts(self, host_list):
        hosts = []
        host_list.return_value = hosts

        res = self.client.get(INDEX_URL)

        host_list.assert_called_once_with(test.IsHttpRequest())
        self.assertTemplateUsed(res, INDEX_TEMPLATE)
        self.assertNoMessages(res)
        self.assertContains(res, 'No items to display')

    @mock.patch.object(blazar_api.client, 'host_list')
    def test_index_error(self, host_list):
        host_list.side_effect = self.exceptions.blazar

        res = self.client.get(INDEX_URL)

        host_list.assert_called_once_with(test.IsHttpRequest())
        self.assertTemplateUsed(res, INDEX_TEMPLATE)
        self.assertMessageCount(res, error=1)

    @mock.patch.object(blazar_api.client, 'host_get')
    def test_host_detail(self, host_get):
        host = self.hosts.get(hypervisor_hostname='compute-1')
        host_get.return_value = host

        res = self.client.get(reverse(DETAIL_URL_BASE, args=[host['id']]))

        host_get.assert_called_once_with(test.IsHttpRequest(), host['id'])
        self.assertTemplateUsed(res, DETAIL_TEMPLATE)
        self.assertContains(res, 'compute-1')
        self.assertContains(res, 'ex1')

    @mock.patch.object(blazar_api.client, 'host_get')
    def test_host_detail_error(self, host_get):
        host_get.side_effect = self.exceptions.blazar

        res = self.client.get(reverse(DETAIL_URL_BASE, args=['invalid']))

        host_get.assert_called_once_with(test.IsHttpRequest(), 'invalid')
        self.assertTemplateNotUsed(res, DETAIL_TEMPLATE)
        self.assertMessageCount(error=1)
        self.assertRedirectsNoFollow(res, INDEX_URL)

    @mock.patch.object(blazar_api.client, 'host_list')
    @mock.patch.object(blazar_api.client, 'host_create')
    @mock.patch.object(api.nova, 'hypervisor_list')
    def test_create_hosts(self, hypervisor_list, host_create, host_list):
        hv_hostnames = [hv.hypervisor_hostname
                        for hv in self.hypervisors.list()]
        calls = []
        for host_name in hv_hostnames:
            calls.append(mock.call(test.IsHttpRequest(), name=host_name))
        form_data = {
            'select_hosts_role_member': hv_hostnames
        }
        host_list.return_value = []
        host_create.return_value = []
        hypervisor_list.return_value = self.hypervisors.list()

        res = self.client.post(CREATE_URL, form_data)
        host_list.assert_called_once_with(test.IsHttpRequest())
        host_create.assert_has_calls(calls)
        self.assertEqual(len(hv_hostnames), host_create.call_count)
        hypervisor_list.assert_called_once_with(test.IsHttpRequest())
        self.assertNoFormErrors(res)
        self.assertMessageCount(success=(len(hv_hostnames) + 1))
        self.assertRedirectsNoFollow(res, INDEX_URL)

    @mock.patch.object(blazar_api.client, 'host_list')
    @mock.patch.object(blazar_api.client, 'host_create')
    @mock.patch.object(api.nova, 'hypervisor_list')
    def test_create_hosts_with_extra_caps(self, hypervisor_list, host_create,
                                          host_list):
        hv_hostnames = [hv.hypervisor_hostname
                        for hv in self.hypervisors.list()]
        calls = []
        for host_name in hv_hostnames:
            calls.append(mock.call(test.IsHttpRequest(),
                                   name=host_name, extracap="strong"))
        form_data = {
            'select_hosts_role_member': hv_hostnames,
            'extra_caps': '{"extracap": "strong"}'
        }
        host_list.return_value = []
        host_create.return_value = []
        hypervisor_list.return_value = self.hypervisors.list()

        res = self.client.post(CREATE_URL, form_data)

        host_list.assert_called_once_with(test.IsHttpRequest())
        host_create.assert_has_calls(calls)
        self.assertEqual(len(hv_hostnames), host_create.call_count)
        hypervisor_list.assert_called_once_with(test.IsHttpRequest())
        self.assertNoFormErrors(res)
        self.assertMessageCount(success=(len(hv_hostnames) + 1))
        self.assertRedirectsNoFollow(res, INDEX_URL)

    @mock.patch.object(blazar_api.client, 'host_get')
    @mock.patch.object(blazar_api.client, 'host_update')
    def test_update_host(self, host_update, host_get):
        host = self.hosts.get(hypervisor_hostname='compute-1')
        form_data = {
            'host_id': host['id'],
            'values': '{"key": "updated"}'
        }
        host_get.return_value = host
        host_update.return_value = []

        res = self.client.post(reverse(UPDATE_URL_BASE, args=[host['id']]),
                               form_data)

        host_get.assert_called_once_with(test.IsHttpRequest(), host['id'])
        host_update.assert_called_once_with(test.IsHttpRequest(),
                                            host_id=host['id'],
                                            values={"key": "updated"})
        self.assertNoFormErrors(res)
        self.assertMessageCount(success=1)
        self.assertRedirectsNoFollow(res, INDEX_URL)

    @mock.patch.object(blazar_api.client, 'host_get')
    @mock.patch.object(blazar_api.client, 'host_update')
    def test_update_host_error(self, host_update, host_get):
        host = self.hosts.get(hypervisor_hostname='compute-1')
        form_data = {
            'host_id': host['id'],
            'values': '{"key": "updated"}'
        }
        host_get.return_value = host
        host_update.side_effect = self.exceptions.blazar

        res = self.client.post(reverse(UPDATE_URL_BASE, args=[host['id']]),
                               form_data)

        host_get.assert_called_once_with(test.IsHttpRequest(), host['id'])
        host_update.assert_called_once_with(test.IsHttpRequest(),
                                            host_id=host['id'],
                                            values={"key": "updated"})
        self.assertNoFormErrors(res)
        self.assertContains(res, 'An error occurred while updating')

    @mock.patch.object(blazar_api.client, 'host_list')
    @mock.patch.object(blazar_api.client, 'host_delete')
    def test_delete_host(self, host_delete, host_list):
        hosts = self.hosts.list()
        host = self.hosts.get(hypervisor_hostname='compute-1')
        action = 'hosts__delete__%s' % host['id']
        form_data = {'action': action}
        host_list.return_value = hosts

        res = self.client.post(INDEX_URL, form_data)

        host_list.assert_called_once_with(test.IsHttpRequest())
        host_delete.assert_called_once_with(test.IsHttpRequest(), host['id'])
        self.assertMessageCount(success=1)
        self.assertRedirectsNoFollow(res, INDEX_URL)

    @mock.patch.object(blazar_api.client, 'host_list')
    @mock.patch.object(blazar_api.client, 'host_delete')
    def test_delete_host_error(self, host_delete, host_list):
        hosts = self.hosts.list()
        host = self.hosts.get(hypervisor_hostname='compute-1')
        action = 'hosts__delete__%s' % host['id']
        form_data = {'action': action}
        host_list.return_value = hosts
        host_delete.side_effect = self.exceptions.blazar

        res = self.client.post(INDEX_URL, form_data)

        host_list.assert_called_once_with(test.IsHttpRequest())
        host_delete.assert_called_once_with(test.IsHttpRequest(), host['id'])
        self.assertMessageCount(error=1)
        self.assertRedirectsNoFollow(res, INDEX_URL)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from django.template import defaultfilters as filters
from django.utils.translation import ugettext_lazy as _
from django.utils.translation import ungettext_lazy
from horizon import tables
from horizon.templatetags import sizeformat

from blazar_dashboard import api


class CreateHosts(tables.LinkAction):
    name = "create"
    verbose_name = _("Create Hosts")
    url = "horizon:admin:hosts:create"
    classes = ("ajax-modal",)
    icon = "plus"


class UpdateHost(tables.LinkAction):
    name = "update"
    verbose_name = _("Update Host")
    url = "horizon:admin:hosts:update"
    classes = ("btn-create", "ajax-modal")


class DeleteHost(tables.DeleteAction):
    name = "delete"
    data_type_singular = _("Host")
    data_type_plural = _("Hosts")
    classes = ('btn-danger', 'btn-terminate')

    @staticmethod
    def action_present(count):
        return ungettext_lazy(
            u"Delete Host",
            u"Delete Hosts",
            count
        )

    @staticmethod
    def action_past(count):
        return ungettext_lazy(
            u"Deleted Host",
            u"Deleted Hosts",
            count
        )

    def delete(self, request, host_id):
        api.client.host_delete(request, host_id)


class HostsTable(tables.DataTable):
    name = tables.Column("hypervisor_hostname", verbose_name=_("Host name"),
                         link="horizon:admin:hosts:detail")
    vcpus = tables.Column("vcpus", verbose_name=_("VCPUs"))
    memory_mb = tables.Column("memory_mb", verbose_name=_("RAM"),
                              filters=(sizeformat.mb_float_format,))
    local_gb = tables.Column("local_gb", verbose_name=_("Local Storage"),
                             filters=(sizeformat.diskgbformat,))
    type = tables.Column("hypervisor_type", verbose_name=_("Hypervisor type"))
    reservable = tables.Column("reservable", verbose_name=_("Reservable"),
                               filters=(filters.yesno, filters.capfirst))

    class Meta(object):
        name = "hosts"
        verbose_name = _("Hosts")
        table_actions = (CreateHosts, DeleteHost,)
        row_actions = (UpdateHost, DeleteHost,)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from django import http
from openstack_dashboard.test import helpers

from blazar_dashboard.test.test_data import utils


class TestCase(helpers.TestCase):
    def _setup_test_data(self):
        super(TestCase, self)._setup_test_data()
        utils.load_test_data(self)


class BaseAdminViewTests(helpers.BaseAdminViewTests):
    def _setup_test_data(self):
        super(BaseAdminViewTests, self)._setup_test_data()
        utils.load_test_data(self)


class IsA(object):
    """Class to compare param is a specified class."""
    def __init__(self, cls):
        self.cls = cls

    def __eq__(self, other):
        return isinstance(other, self.cls)


class IsHttpRequest(IsA):
    """Class to compare param is django.http.HttpRequest."""
    def __init__(self):
        super(IsHttpRequest, self).__init__(http.HttpRequest)



# Copyright 2014 Intel Corporation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import datetime
import json
import logging
import re

from django.utils.translation import ugettext_lazy as _
from horizon import exceptions
from horizon import forms
from horizon import messages
from pytz import timezone

from blazar_dashboard import api

LOG = logging.getLogger(__name__)


class CreateForm(forms.SelfHandlingForm):
    # General fields
    name = forms.CharField(
        label=_("Lease Name"),
        required=True,
        max_length=80
    )
    start_date = forms.DateTimeField(
        label=_("Start Date"),
        required=False,
        help_text=_('Enter YYYY-MM-DD HH:MM or blank for now'),
        input_formats=['%Y-%m-%d %H:%M'],
        widget=forms.DateTimeInput(
            attrs={'placeholder': 'YYYY-MM-DD HH:MM (blank for now)'})
    )
    end_date = forms.DateTimeField(
        label=_("End Date"),
        required=False,
        help_text=_('Enter YYYY-MM-DD HH:MM or blank for Start Date + 24h'),
        input_formats=['%Y-%m-%d %H:%M'],
        widget=forms.DateTimeInput(
            attrs={'placeholder': 'YYYY-MM-DD HH:MM (blank for Start Date + '
                                  '24h)'})
    )
    resource_type = forms.ChoiceField(
        label=_("Resource Type"),
        required=True,
        choices=(
            ('host', _('Physical Host')),
            ('instance', _('Virtual Instance'))
        ),
        widget=forms.ThemableSelectWidget(attrs={
            'class': 'switchable',
            'data-slug': 'source'}))

    # Fields for host reservation
    min_hosts = forms.IntegerField(
        label=_('Minimum Number of Hosts'),
        required=False,
        help_text=_('Enter the minimum number of hosts to reserve.'),
        min_value=1,
        initial=1,
        widget=forms.NumberInput(attrs={
            'class': 'switched',
            'data-switch-on': 'source',
            'data-source-host': _('Minimum Number of Hosts')})
    )
    max_hosts = forms.IntegerField(
        label=_('Maximum Number of Hosts'),
        required=False,
        help_text=_('Enter the maximum number of hosts to reserve.'),
        min_value=1,
        initial=1,
        widget=forms.NumberInput(attrs={
            'class': 'switched',
            'data-switch-on': 'source',
            'data-source-host': _('Maximum Number of Hosts')})
    )
    hypervisor_properties = forms.CharField(
        label=_("Hypervisor Properties"),
        required=False,
        help_text=_('Enter properties of a hypervisor to reserve.'),
        max_length=255,
        widget=forms.TextInput(attrs={
            'class': 'switched',
            'data-switch-on': 'source',
            'data-source-host': _('Hypervisor Properties'),
            'placeholder': 'e.g. [">=", "$vcpus", "2"]'})
    )

    # Fields for instance reservation
    amount = forms.IntegerField(
        label=_('Instance Count'),
        required=False,
        help_text=_('Enter the number of instances to reserve.'),
        min_value=1,
        initial=1,
        widget=forms.NumberInput(attrs={
            'class': 'switched',
            'data-switch-on': 'source',
            'data-source-instance': _('Instance Count')})
    )
    vcpus = forms.IntegerField(
        label=_('Number of VCPUs'),
        required=False,
        help_text=_('Enter the number of VCPUs per instance.'),
        min_value=1,
        initial=1,
        widget=forms.NumberInput(attrs={
            'class': 'switched',
            'data-switch-on': 'source',
            'data-source-instance': _('Number of VCPUs')})
    )
    memory_mb = forms.IntegerField(
        label=_('RAM (MB)'),
        required=False,
        help_text=_('Enter the size of RAM (MB) per instance'),
        min_value=1,
        initial=1,
        widget=forms.NumberInput(attrs={
            'class': 'switched',
            'data-switch-on': 'source',
            'data-source-instance': _('RAM (MB)')})
    )
    disk_gb = forms.IntegerField(
        label=_('Root Disk (GB)'),
        required=False,
        help_text=_('Enter the root disk size (GB) per instance'),
        min_value=0,
        initial=0,
        widget=forms.NumberInput(attrs={
            'class': 'switched',
            'data-switch-on': 'source',
            'data-source-instance': _('Root Disk (GB)')})
    )
    affinity = forms.ChoiceField(
        label=_("Affinity Rule"),
        required=False,
        choices=(
            (None, _('None')),
            (True, _('Affinity')),
            (False, _('Anti-Affinity')),
        ),
        initial=None,
        widget=forms.ThemableSelectWidget(attrs={
            'class': 'switched',
            'data-switch-on': 'source',
            'data-source-instance': _('Affinity Rule')})
    )

    # Fields for both of host and instance reservations
    resource_properties = forms.CharField(
        label=_("Resource Properties"),
        required=False,
        help_text=_('Enter properties of a resource to reserve.'),
        max_length=255,
        widget=forms.TextInput(attrs={
            'placeholder': 'e.g. ["==", "$extra_key", "extra_value"]'})
    )

    def handle(self, request, data):
        if data['resource_type'] == 'host':
            reservations = [
                {
                    'resource_type': 'physical:host',
                    'min': data['min_hosts'],
                    'max': data['max_hosts'],
                    'hypervisor_properties': (data['hypervisor_properties'] or
                                              ''),
                    'resource_properties': data['resource_properties'] or ''
                }
            ]
        elif data['resource_type'] == 'instance':
            reservations = [
                {
                    'resource_type': 'virtual:instance',
                    'amount': data['amount'],
                    'vcpus': data['vcpus'],
                    'memory_mb': data['memory_mb'],
                    'disk_gb': data['disk_gb'],
                    'affinity': data['affinity'],
                    'resource_properties': data['resource_properties'] or ''
                }
            ]

        events = []

        try:
            api.client.lease_create(
                request, data['name'],
                data['start_date'].strftime('%Y-%m-%d %H:%M'),
                data['end_date'].strftime('%Y-%m-%d %H:%M'),
                reservations, events)
            messages.success(request, _('Lease %s was successfully '
                                        'created.') % data['name'])
            return True
        except Exception as e:
            LOG.error('Error submitting lease: %s', e)
            exceptions.handle(request,
                              message='An error occurred while creating this '
                                      'lease: %s.' % e)

    def clean(self):
        cleaned_data = super(CreateForm, self).clean()
        local = timezone(self.request.session.get(
            'django_timezone',
            self.request.COOKIES.get('django_timezone', 'UTC')))

        if cleaned_data['start_date']:
            cleaned_data['start_date'] = local.localize(
                cleaned_data['start_date'].replace(tzinfo=None)
            ).astimezone(timezone('UTC'))
        else:
            cleaned_data['start_date'] = datetime.datetime.utcnow()
        if cleaned_data['end_date']:
            cleaned_data['end_date'] = local.localize(
                cleaned_data['end_date'].replace(tzinfo=None)
            ).astimezone(timezone('UTC'))
        else:
            cleaned_data['end_date'] = (cleaned_data['start_date'] +
                                        datetime.timedelta(days=1))


class UpdateForm(forms.SelfHandlingForm):

    class Meta(object):
        name = _('Update Lease Parameters')

    lease_id = forms.CharField(
        label=_('Lease ID'), widget=forms.widgets.HiddenInput, required=True)
    lease_name = forms.CharField(
        label=_('Lease name'), widget=forms.TextInput(), required=False)
    start_time = forms.CharField(
        label=_('Start time'),
        widget=forms.TextInput(
            attrs={'placeholder': _('Valid suffix are d/h/m (e.g. +1h)')}),
        required=False)
    end_time = forms.CharField(
        label=_('End time'),
        widget=forms.TextInput(
            attrs={'placeholder': _('Valid suffix are d/h/m (e.g. +1h)')}),
        required=False)
    reservations = forms.CharField(
        label=_("Reservation values to update"),
        help_text=_('Enter reservation values to update as JSON'),
        widget=forms.Textarea(
            attrs={'rows': 8,
                   'placeholder':
                   'e.g.\n'
                   '[\n'
                   '    {\n'
                   '        "id": "087bc740-6d2d-410b-9d47-c7b2b55a9d36",\n'
                   '        "max": 3\n'
                   '    }\n'
                   ']'}),
        max_length=511,
        required=False)

    def __init__(self, request, *args, **kwargs):
        super(UpdateForm, self).__init__(request, *args, **kwargs)
        for reservation in kwargs['initial']['lease'].reservations:
            if reservation['resource_type'] == 'virtual:instance':
                # Hide the start/end_time because they cannot be updated if at
                # least one virtual:instance reservation is included.
                # TODO(hiro-kobayashi) remove this part if virtual:instance
                # reservation gets to support update of the start/end_time.
                del self.fields['start_time']
                del self.fields['end_time']
                del self.fields['reservations']
                return

    def handle(self, request, data):
        lease_id = data.get('lease_id')

        fields = {}

        lease_name = data.get('lease_name', None)
        if lease_name:
            fields['name'] = lease_name

        start_time = data.get('start_time', None)
        end_time = data.get('end_time', None)
        if start_time:
            if start_time[0] == '+':
                fields['defer_by'] = start_time[1:]
            elif start_time[0] == '-':
                fields['advance_by'] = start_time[1:]
        if end_time:
            if end_time[0] == '+':
                fields['prolong_for'] = end_time[1:]
            elif end_time[0] == '-':
                fields['reduce_by'] = end_time[1:]

        reservations = data.get('reservations', None)
        if reservations:
            fields['reservations'] = reservations

        try:
            api.client.lease_update(self.request, lease_id=lease_id, **fields)
            messages.success(request, _("Lease update started."))
            return True
        except Exception as e:
            LOG.error('Error updating lease: %s', e)
            exceptions.handle(request,
                              message="An error occurred while updating this"
                                      " lease: %s." % e)

    def clean(self):
        cleaned_data = super(UpdateForm, self).clean()

        lease_name = cleaned_data.get("lease_name", None)
        start_time = cleaned_data.get("start_time", None)
        end_time = cleaned_data.get("end_time", None)
        reservations = cleaned_data.get("reservations", None)

        if start_time:
            valid = re.match(r'^[+-]\d+[dhm]$', start_time)
            if not valid:
                raise forms.ValidationError("The start/end time must be "
                                            "a form of +/- number d/h/m. "
                                            "(e.g. +1h)")

        if end_time:
            valid = re.match(r'^[+-]\d+[dhm]$', end_time)
            if not valid:
                raise forms.ValidationError("The start/end time must be "
                                            "a form of +/- number d/h/m. "
                                            "(e.g. +1h)")

        if reservations:
            try:
                reservations = json.loads(reservations)
                cleaned_data['reservations'] = reservations
            except json.JSONDecodeError:
                raise forms.ValidationError(
                    _('Reservation values must be written in JSON')
                )

        if not (lease_name or start_time or end_time or reservations):
            raise forms.ValidationError("Nothing to update.")



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from datetime import datetime
from itertools import chain
import json
import logging
from pytz import UTC

from blazar_dashboard import conf
from horizon import exceptions
from horizon.utils.memoized import memoized
from openstack_dashboard.api import base

from blazarclient import client as blazar_client


LOG = logging.getLogger(__name__)


class Lease(base.APIDictWrapper):
    """Represents one Blazar lease."""
    _attrs = ['id', 'name', 'start_date', 'end_date', 'user_id', 'project_id',
              'before_end_date', 'status', 'degraded']

    def __init__(self, apiresource):
        super(Lease, self).__init__(apiresource)


class Host(base.APIDictWrapper):
    """Represents one Blazar host."""

    _attrs = ['id', 'hypervisor_hostname', 'hypervisor_type',
              'hypervisor_version', 'vcpus', 'cpu_info', 'memory_mb',
              'local_gb', 'status', 'created_at', 'updated_at',
              'service_name', 'trust_id', 'reservable']

    def __init__(self, apiresource):
        super(Host, self).__init__(apiresource)

    def cpu_info_dict(self):
        cpu_info_dict = getattr(self, 'cpu_info', '{}')
        if not cpu_info_dict:
            cpu_info_dict = '{}'
        return json.loads(cpu_info_dict)

    def extra_capabilities(self):
        excaps = {}
        for k, v in self._apidict.items():
            if k not in self._attrs:
                excaps[k] = v
        return excaps


class Allocation(base.APIDictWrapper):

    _attrs = ['resource_id', 'reservations']


@memoized
def blazarclient(request):
    try:
        api_url = base.url_for(request, 'reservation')
    except exceptions.ServiceCatalogException:
        LOG.debug('No Reservation service is configured.')
        return None

    LOG.debug('blazarclient connection created using the token "%s" and url'
              '"%s"' % (request.user.token.id, api_url))
    return blazar_client.Client(
        blazar_url=api_url,
        auth_token=request.user.token.id)


def lease_list(request):
    """List the leases."""
    leases = blazarclient(request).lease.list()
    return [Lease(l) for l in leases]


def lease_get(request, lease_id):
    """Get a lease."""
    lease = blazarclient(request).lease.get(lease_id)
    return Lease(lease)


def lease_create(request, name, start, end, reservations, events):
    """Create a lease."""
    lease = blazarclient(request).lease.create(
        name, start, end, reservations, events)
    return Lease(lease)


def lease_update(request, lease_id, **kwargs):
    """Update a lease."""
    lease = blazarclient(request).lease.update(lease_id, **kwargs)
    return Lease(lease)


def lease_delete(request, lease_id):
    """Delete a lease."""
    blazarclient(request).lease.delete(lease_id)


def host_list(request):
    """List hosts."""
    hosts = blazarclient(request).host.list()
    return [Host(h) for h in hosts]


def host_get(request, host_id):
    """Get a host."""
    host = blazarclient(request).host.get(host_id)
    return Host(host)


def host_create(request, name, **kwargs):
    """Create a host."""
    host = blazarclient(request).host.create(name, **kwargs)
    return Host(host)


def host_update(request, host_id, values):
    """Update a host."""
    host = blazarclient(request).host.update(host_id, values)
    return Host(host)


def host_delete(request, host_id):
    """Delete a host."""
    blazarclient(request).host.delete(host_id)


def host_allocations_list(request):
    """List allocations for all hosts."""
    request_manager = blazarclient(request).host.request_manager
    resp, body = request_manager.get('/os-hosts/allocations')
    allocations = body['allocations']
    return [Allocation(a) for a in allocations]


def reservation_calendar(request):
    """Return a list of all scheduled leases."""

    def compute_host2dict(h):
        dictionary = dict(
            hypervisor_hostname=h.hypervisor_hostname, vcpus=h.vcpus,
            memory_mb=h.memory_mb, local_gb=h.local_gb, cpu_info=h.cpu_info,
            hypervisor_type=h.hypervisor_type,)
        # Ensure config attribute is copied over
        calendar_attribute = conf.host_reservation.get('calendar_attribute')
        dictionary[calendar_attribute] = (
            h[calendar_attribute]
        )
        return dictionary

    # NOTE: This filters by reservable hosts
    hosts_by_id = {h.id: h for h in host_list(request) if h.reservable}

    def host_reservation_dict(reservation, resource_id):
        host_reservation = dict(
            start_date=_parse_api_datestr(reservation['start_date']),
            end_date=_parse_api_datestr(reservation['end_date']),
            reservation_id=reservation['id'],
        )
        calendar_attribute = conf.host_reservation.get('calendar_attribute')
        host_reservation[calendar_attribute] = (
            hosts_by_id[resource_id][calendar_attribute]
        )

        return {k: v for k, v in host_reservation.items() if v is not None}

    host_reservations = [
        [host_reservation_dict(r, alloc.resource_id)
            for r in alloc.reservations
            if alloc.resource_id in hosts_by_id]
        for alloc in host_allocations_list(request)]

    compute_hosts = [compute_host2dict(h) for h in hosts_by_id.values()]

    return compute_hosts, list(chain(*host_reservations))


def _parse_api_datestr(datestr):
    if datestr is None:
        return datestr

    dateobj = datetime.strptime(datestr, "%Y-%m-%dT%H:%M:%S.%f")

    return dateobj.replace(tzinfo=UTC)



# Copyright 2014 Intel Corporation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from datetime import datetime
from functools import partial

from django.template import defaultfilters as django_filters
from django.utils.translation import ugettext_lazy as _
from django.utils.translation import ungettext_lazy
from horizon import tables
from horizon.utils import filters
import pytz

from blazar_dashboard import api
from blazar_dashboard import conf


class CreateLease(tables.LinkAction):
    name = "create"
    verbose_name = _("Create Lease")
    url = "horizon:project:leases:create"
    classes = ("ajax-modal",)
    icon = "plus"


class UpdateLease(tables.LinkAction):
    name = "update"
    verbose_name = _("Update Lease")
    url = "horizon:project:leases:update"
    classes = ("btn-create", "ajax-modal")

    def allowed(self, request, lease):
        if datetime.strptime(lease.end_date, '%Y-%m-%dT%H:%M:%S.%f').\
                replace(tzinfo=pytz.utc) > datetime.now(pytz.utc):
            return True
        return False


class ViewHostReservationCalendar(tables.LinkAction):
    name = "calendar"
    verbose_name = _("Host Calendar")
    url = "calendar/host/"
    classes = ("btn-default", )
    icon = "calendar"


class DeleteLease(tables.DeleteAction):
    name = "delete"
    data_type_singular = _("Lease")
    data_type_plural = _("Leases")
    classes = ('btn-danger', 'btn-terminate')

    @staticmethod
    def action_present(count):
        return ungettext_lazy(
            u"Delete Lease",
            u"Delete Leases",
            count
        )

    @staticmethod
    def action_past(count):
        return ungettext_lazy(
            u"Deleted Lease",
            u"Deleted Leases",
            count
        )

    def delete(self, request, lease_id):
        api.client.lease_delete(request, lease_id)


class LeasesTable(tables.DataTable):
    name = tables.Column("name", verbose_name=_("Lease name"),
                         link="horizon:project:leases:detail",)
    start_date = tables.Column("start_date", verbose_name=_("Start date"),
                               filters=(filters.parse_isotime,
                                        partial(django_filters.date,
                                                arg='Y-m-d H:i T')),)
    end_date = tables.Column("end_date", verbose_name=_("End date"),
                             filters=(filters.parse_isotime,
                                      partial(django_filters.date,
                                              arg='Y-m-d H:i T')),)
    status = tables.Column("status", verbose_name=_("Status"),)
    degraded = tables.Column("degraded", verbose_name=_("Degraded"),
                             filters=(django_filters.yesno,
                                      django_filters.capfirst),)

    class Meta(object):
        name = "leases"
        verbose_name = _("Leases")
        table_actions = [CreateLease, DeleteLease, ]
        if conf.host_reservation.get('enabled'):
            table_actions.insert(0, ViewHostReservationCalendar)
        row_actions = (UpdateLease, DeleteLease, )
